from .functional import direct_norm, project_norm, get_normed_weights
from .LipschitzMonotonicNetwork import (
    MonotonicWrapper,
    MonotonicLayer,
    LipschitzLinear,
    RMSNorm,
    SigmaNet
)
from .group import GroupSort

__all__ = [
    "direct_norm",
    "project_norm",
    "MonotonicWrapper",
    "GroupSort",
    "get_normed_weights",
    "MonotonicLayer",
    "RMSNorm",
    "LipschitzLinear",
    "SigmaNet",
]



from torch import nn
import torch
from torch.nn.utils.parametrize import register_parametrization
import typing as t

kinds = [
    "one",  # |W|_1 constraint
    "inf",  # |W|_inf constraint
    "one-inf",  # |W|_1,inf constraint
    "two-inf",  # |W|_2,inf constraint
]


def direct_norm(
    layer: nn.Linear,
    kind: str = "one",
    always_norm: bool = False,
    max_norm: t.Optional[float] = None,
    parameter_name: str = "weight",
    vectorwise: bool = True,
) -> nn.Linear:
    """
    Constrain the norm of a layer's weight matrix. This function is meant to be
    used as a wrapper around `torch.nn.Linear`. It adds a hook to the forward pass 
    to constrain the norm of the weight matrix. The hook is registered as a forward pre-hook.
    This means that the weights are normalized before the forward pass and that the gradients
    are backpropagated through normalization back to the original weights.


    Args:
        layer (nn.Linear): The layer to constrain.
        kind (str, optional): The kind of norm to constrain options are
             [
                "one",  # |W|_1 constraint
                "inf",  # |W|_inf constraint
                "one-inf",  # |W|_1,inf constraint
                "two-inf",  # |W|_2,inf constraint
            ]
            Defaults to "one".
        always_norm (bool, optional): Always normalize the weights to max_norm or allow them 
            to have a norm smaller than 1. Defaults to False.
        max_norm (float, optional): The maximum norm to constrain the weights to. Defaults to 1. 
        parameter_name (str, optional): Name of the weight matrix in the layer. Defaults to "weight".
        vectorwise (bool, optional): Normalize the matrix vectors instead of the matrix itself. 
            The vector norm is a weaker constraint and gives better results. See the paper for details.
            Defaults to True.
    Raises:
        ValueError: If kind is not one of the options.

    Returns:
        nn.Linear: The layer with the hook registered. Matrix multiplication will be done with 
            normalized version of the weights.
    """
    if kind not in kinds:
        raise ValueError(f"kind {kind} not recognized. Choose one of {kinds}")

    class Normalize(nn.Module):
      def forward(self, W):
        return get_normed_weights(W, kind, always_norm, max_norm, vectorwise)

    register_parametrization(layer, parameter_name, Normalize())

    return layer


def project_norm(
    layer: nn.Linear,
    kind: str = "one",
    always_norm: bool = True,
    max_norm: t.Optional[float] = None,
    parameter_name: str = "weight",
    vectorwise: bool = True,
) -> nn.Linear:
    """
    Constrain the norm of a layer's weight matrix by projecting the weight matrix to the correct norm.
    This function is meant to be used as a wrapper around `torch.nn.Linear`. It adds a hook to the forward pass
    which projects to the current weight matrix to a normalized matrix but the gradients are not backpropagated 
    through normalization. 

    Args:
        layer (nn.Linear): The layer to constrain.
        kind (str, optional): The kind of norm to constrain options are
             [
                "one",  # |W|_1 constraint
                "inf",  # |W|_inf constraint
                "one-inf",  # |W|_1,inf constraint
                "two-inf",  # |W|_2,inf constraint
            ]
            Defaults to "one".
        always_norm (bool, optional): Always normalize the weights to max_norm or allow them 
            to have a norm smaller than 1. Defaults to False.
        max_norm (float, optional): The maximum norm to constrain the weights to. Defaults to 1. 
        parameter_name (str, optional): Name of the weight matrix in the layer. Defaults to "weight".
        vectorwise (bool, optional): Normalize the matrix vectors instead of the matrix itself. 
            The vector norm is a weaker constraint and gives better results. See the paper for details.
            Defaults to True.
    Raises:
        ValueError: If kind is not one of the options.

    Returns:
        nn.Linear: The layer with the hook registered. Matrix multiplication will be done with 
            normalized version of the weights.
    """
    if kind not in kinds:
        raise ValueError(f"kind {kind} not recognized. Choose one of {kinds}")

    @torch.no_grad()
    def normalize_weight(layer: nn.Linear, _) -> None:
        weight = getattr(layer, parameter_name).detach()
        weight = get_normed_weights(weight, kind, always_norm, max_norm, vectorwise)
        getattr(layer, parameter_name).data.copy_(weight)

    layer.register_forward_pre_hook(normalize_weight)
    return layer


def get_normed_weights(
    weight: torch.Tensor,
    kind: str,
    always_norm: bool,
    max_norm: t.Optional[float],
    vectorwise: bool,
) -> torch.Tensor:
    """
    Normalize weight matrix to a given norm.

    Args:
        weight (torch.Tensor): The weight matrix to normalize.
        kind (str): The kind of norm to constrain options are
                [
                    "one",  # |W|_1 constraint
                    "inf",  # |W|_inf constraint
                    "one-inf",  # |W|_1,inf constraint
                    "two-inf",  # |W|_2,inf constraint
                ]
        always_norm (bool): Always normalize the weights to max_norm or allow them
            to have a norm smaller than 1.
        max_norm (t.Optional[float]): The maximum norm to constrain the weights to.
        vectorwise (bool): Normalize the matrix vectors instead of the matrix itself.

    Returns:
        torch.Tensor: The normalized weight matrix.
    """
    if kind == "one":
        norms = weight.abs().sum(axis=0)
    elif kind == "inf":
        norms = weight.abs().sum(axis=1, keepdim=True)
    elif kind == "one-inf":
        norms = weight.abs()
    elif kind == "two-inf":
        norms = torch.norm(weight, p=2, dim=1, keepdim=True)
    if not vectorwise:
        norms = norms.max()

    max_norm = max_norm or 1

    if not always_norm:
        norms = torch.max(torch.ones_like(norms), norms / max_norm)
    else:
        norms = norms / max_norm

    weight = weight / torch.max(norms, torch.ones_like(norms)*1e-10)

    return weight



import torch
import typing as T
from .functional import direct_norm
import warnings
from torch.nn import Module, Parameter, Linear


class MonotonicWrapper(Module):
    def __init__(
        self,
        lipschitz_module: Module,  # Must already be lipschitz
        lipschitz_const: float = 1.0,
        monotonic_constraints: T.Optional[T.Iterable] = None,
    ):
        """This is a wrapper around a module with a lipschitz_const lipschitz constant. It
            adds a term to the output of the module which enforces monotonicity constraints
            given by monotonic_constraints.
            Returns a module which is monotonic and Lipschitz with constant lipschitz_const.

        Args:
            lipschitz_module (torch.nn.Module): Lipschitz-constrained nn.Module with Lipschitz
                constant lipschitz_const.
            lipschitz_const (float, optional): Lipschitz constant of the network given in nn.
                Defaults to 1.
            monotonic_constraints (torch.Tensor, optional): Monotonicity constraints on the inputs.
                If None, all inputs are assumed to be monotonically increasing.
                If monotonic_constraints[i] = 1, the i-th input is constrained to be non-decreasing.
                If monotonic_constraints[i] = -1, the i-th input is constrained to be
                non-increasing.
                If monotonic_constraints[i] = 0, the i-th input is unconstrained.

                If a 1D tensor, the same constraint is applied to all ouputs.
                If a 2D tensor, the (i, j)-th element specifies the constraint on the j-th output
                with respect to the i-th input.
                Default: ``None``
        """
        super().__init__()
        self.nn = lipschitz_module
        self.register_buffer(
            "lipschitz_const", torch.Tensor([lipschitz_const])
        )
        if monotonic_constraints is None:
            monotonic_constraints = [1]
        monotonic_constraints = torch.Tensor(monotonic_constraints)
        if monotonic_constraints.ndim == 1:
            monotonic_constraints = monotonic_constraints.unsqueeze(-1)
        self.register_buffer("monotonic_constraints", monotonic_constraints)

    def forward(self, x: torch.Tensor):
        mc = self.monotonic_constraints.expand(x.shape[1], -1)  # type: ignore
        residual = self.lipschitz_const * x @ mc
        return self.nn(x) + residual


class LipschitzLinear(Linear):
    """
    A linear layer with a Lipschitz constraint on its weights.

    Args:
        in_features (int): Size of each input sample.
        out_features (int): Size of each output sample.
        bias (bool, optional): If set to ``False``, the layer will not learn an additive bias.
            Default: ``True``
        lipschitz_const (float, optional): Lipschitz constant of the layer.
            Default: ``1``
        kind (str, optional): Type of Lipschitz constraint to enforce. Options are
            - "one",  # |W|_1 constraint
            - "inf",  # |W|_inf constraint
            - "one-inf",  # |W|_1,inf constraint
            - "two-inf",  # |W|_2,inf constraint
            Will be passed to :func:`direct_norm`, see its documentation for more details.
            Defaults to "one".
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        lipschitz_const: float = 1.0,
        kind: str = "one-inf",
    ):
        super().__init__(in_features, out_features, bias=bias)
        self.register_buffer(
            "lipschitz_const", torch.Tensor([lipschitz_const])
        )
        # Directly enforce Lipschitz constraint
        self = direct_norm(self, max_norm=lipschitz_const, kind=kind)


class MonotonicLayer(LipschitzLinear):
    """
    A linear layer with a Lipschitz constraint on its weights and monotonicity constraints.

    Args:
        in_features (int): Size of each input sample.
        out_features (int): Size of each output sample.
        bias (bool, optional): If set to ``False``, the layer will not learn an additive bias.
            Default: ``True``
        lipschitz_const (float, optional): Lipschitz constant of the layer.
            Default: ``1``
        monotonic_constraints (torch.Tensor, optional): Monotonicity constraints on the inputs.
            If None, all inputs are assumed to be monotonically increasing.
            If monotonic_constraints[i] = 1, the i-th input is constrained to be non-decreasing.
            If monotonic_constraints[i] = -1, the i-th input is constrained to be non-increasing.
            If monotonic_constraints[i] = 0, the i-th input is unconstrained.

            If a 1D tensor, the same constraint is applied to all ouputs.
            If a 2D tensor, the (i, j)-th element specifies the constraint on the j-th output
            with respect to the i-th input.
            Default: ``None``
        kind (str, optional): Type of Lipschitz constraint to enforce. Default: ``"one-inf"``
            Options are
            - "one",  # |W|_1 constraint
            - "inf",  # |W|_inf constraint
            - "one-inf",  # |W|_1,inf constraint
            - "two-inf",  # |W|_2,inf constraint
            Will be passed to :func:`direct_norm`, see its documentation for more details.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        lipschitz_const: float = 1.0,
        monotonic_constraints: T.Optional[T.Iterable] = None,
        kind: str = "one-inf",
    ):
        super().__init__(
            in_features, out_features, bias, lipschitz_const, kind
        )
        self.register_buffer(
            "lipschitz_const", torch.tensor([lipschitz_const])
        )
        if monotonic_constraints is None:
            monotonic_constraints = torch.ones(in_features)
        else:
            monotonic_constraints = torch.Tensor(monotonic_constraints)

        if monotonic_constraints.shape not in [
            (in_features, out_features),
            (in_features, 1),
            (in_features,),
        ]:
            raise ValueError(
                "monotonic_constraints must be of shape (in_features, out_features),"
                " ,(in_features, 1), or (in_features,)"
                f" got {monotonic_constraints.shape}"
            )

        if monotonic_constraints.ndim == 1:
            monotonic_constraints = monotonic_constraints.unsqueeze(-1)
        self.register_buffer("monotonic_constraints", monotonic_constraints)

    def forward(self, x: torch.Tensor):
        residual = self.lipschitz_const * x @ self.monotonic_constraints
        x = torch.nn.functional.linear(x, self.weight, self.bias)
        return (x + residual) / 2


class RMSNorm(Module):
    def __init__(
        self, norm_shape: T.Union[T.Iterable, int], affine: bool = True
    ):
        super().__init__()
        self.register_buffer("norm_shape", torch.tensor(norm_shape))
        weights = torch.ones(norm_shape) / self.norm_shape.sqrt()  # type: ignore
        self.register_parameter("weight", Parameter(weights, affine))
        bias = torch.zeros(norm_shape)  # type: ignore
        self.register_parameter("bias", Parameter(bias, affine))

    def forward(self, x):
        rms = torch.sqrt(torch.mean(x**2, dim=-1, keepdim=True)).clip(min=1)
        return (x / rms) * self.weight.pow(2) + self.bias  # type: ignore


class SigmaNet(MonotonicWrapper):
    def __init__(
        self,
        nn: Module,  # Must already be sigma lipschitz
        sigma: float,
        monotonic_constraints: T.Optional[T.Iterable] = None,
    ):
        """Implementation of a monotone network with a sigma lipschitz constraint.

        Args:
            nn (torch.nn.Module): Lipschitz-constrained network with Lipschitz
                constant sigma.
            monotonic_constraints (T.Optional[T.Iterable], optional): Iterable of the
                monotonic features. For example, if a network
                which takes a vector of size 3 is meant to be monotonic in the last
                feature only, then monotonic_constraints should be [0, 0, 1].
                Defaults to all features (i.e. a vector of ones everywhere).
                Montonically deacreasing features should have value -1 instead of 1.
            sigma (float, optional): Lipschitz constant of the network given in nn. Defaults to 1.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("default", DeprecationWarning)
            warnings.warn(
                "SigmaNet is deprecated, use MonotonicWrapper instead",
                DeprecationWarning,
            )
        super().__init__(nn, sigma, monotonic_constraints)



import torch


class GroupSort(torch.nn.Module):
    def __init__(self, n_groups: int, axis: int = -1):
        super(GroupSort, self).__init__()
        self.n_groups = n_groups
        self.axis = axis

    def forward(self, x: torch.Tensor):
        return group_sort(x, self.n_groups, self.axis)

    def extra_repr(self):
        return f"num_groups: {self.n_groups}"


def get_sorting_shape(x: torch.Tensor, n_groups: int, axis: int = -1) -> list:
    shape = list(x.shape)
    num_features = shape[axis]
    if num_features % n_groups:
        raise ValueError(
            f"number of features({num_features}) needs to be a multiple of n_groups({n_groups})"
        )
    shape[axis] = -1
    n_per_group = num_features // n_groups
    if axis == -1:
        shape.append(n_per_group)
    else:
        shape.insert(axis + 1, n_per_group)
    return shape


def group_sort(x: torch.Tensor, n_groups: int, axis: int = -1) -> torch.Tensor:
    if x.shape[0] == 0:
        return x
    size = get_sorting_shape(x, n_groups, axis)
    grouped_x = x.view(*size)
    sort_dim = axis if axis == -1 else axis + 1
    sorted_grouped_x, _ = grouped_x.sort(dim=sort_dim)
    sorted_x = sorted_grouped_x.view(*x.shape)
    return sorted_x



# Bug fix for project_norm
import unittest
from .functional import direct_norm, project_norm
from .group import GroupSort
from .LipschitzMonotonicNetwork import LipschitzLinear
import torch


class TestNorms(unittest.TestCase):
    def test_no_norm(self):
        loss = train(lambda x, **kwargs: x)
        self.assertAlmostEqual(loss, 0.0, places=4)

    def test_project_norm(self):
        loss = train(project_norm)
        self.assertAlmostEqual(loss, 0.0, places=4)

    def test_direct_norm(self):
        loss = train(direct_norm)
        self.assertAlmostEqual(loss, 0.0, places=4)


def train(norm):
    model = torch.nn.Sequential(
        norm(torch.nn.Linear(1, 32), kind="two-inf", always_norm=False),
        GroupSort(2),
        norm(torch.nn.Linear(32, 32), kind="inf", always_norm=False),
        GroupSort(2),
        norm(torch.nn.Linear(32, 1), kind="inf", always_norm=False),
    )
    # define training target
    # fit f(x) = x^2 / 2
    x = torch.linspace(-1, 1, 100).view(-1, 1)
    y = x**2 / 2

    optim = torch.optim.Adam(model.parameters(), lr=0.001)

    for i in range(2000):
        loss = torch.mean((model(x) - y) ** 2)
        loss.backward()
        optim.step()
        optim.zero_grad()
    return loss.item()  # type: ignore


if __name__ == "__main__":
    torch.manual_seed(0)
    # run the tests
    unittest.main()



from setuptools import setup

setup(
    name="monotonicnetworks",
    version="1.5.0",
    packages=["monotonicnetworks"],
    url="",
    license="MIT",
    author="Ouail Kitouni, Niklas Nolte",
    author_email="kitouni@mit.edu, nolte@meta.com",
    description=(
        "Pytorch implementation of constrained weight operator norms for robustness "
        + "and monotonicity in neural networks."
    ),
)
